/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

`define TIM0_FUNC_MINIMUM

// Module Name: rtavr_ior_timer0
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     49         49   
// Number of 4 input LUTs:         69         73
// Number of occupied Slices:      56         57
// Total Number of 4 input LUTs:   78         82 
// Number of bonded IOBs :         28
//   IOB Flip Flops :              -
// Number of BUFGMUXs :             1
// Number of RAMB16BWEs :          -

module rtavr_ior_timer0(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO

		, input WR,RD
		, input [5:0] ADDR

		// external I/O
		, output TIM0_OVF
		, output TIM0_COMPA	
		, output TIM0_COMPB
		);
  parameter BASE_ADDR = 6'h15; // OCR0B 0x15
  reg  [9:0] r_prescaler;
  reg  [7:0] r_tcnt0;
  reg  [7:0] r_ocr0b;
  reg  [7:0] r_ocr0a;
  reg  r_tim0_compa;
  reg  r_tim0_compb;
  reg  r_tim0_ovf;
 
  assign TIM0_OVF = r_tim0_ovf;
  assign TIM0_COMPA = r_tim0_compa;
  assign TIM0_COMPB = r_tim0_compb;


  wire OE_ocr0b = RD & (ADDR == BASE_ADDR);
  wire WE_ocr0b = WR & (ADDR == BASE_ADDR);
  wire OE_ocr0a = RD & (ADDR == (BASE_ADDR + 1));
  wire WE_ocr0a = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_tcnt0 = RD & (ADDR == (BASE_ADDR + 2));
  wire WE_tcnt0 = WR & (ADDR == (BASE_ADDR + 2));
  wire OE_tccr0b = RD & (ADDR == (BASE_ADDR + 3));
  wire WE_tccr0b = WR & (ADDR == (BASE_ADDR + 3));
  wire OE_tccr0a = RD & (ADDR == (BASE_ADDR + 4));
  wire WE_tccr0a = WR & (ADDR == (BASE_ADDR + 4));

  reg [2:0] cs0;
  reg [2:0] wgm0;
  reg [1:0] com0a;
  reg [1:0] com0b;

  reg tsm;
  reg psr;
`ifdef TIM0_FUNC_MINIMUM
`else
  reg foc0b;
  reg foc0a;
`endif
  wire  [7:0] i_tccr0a = {com0a, com0b, wgm0[2], cs0};
  wire  [7:0] i_tccr0b = {2'b0, tsm, 3'b0 , wgm0[1:0]};

  wire v_psr = (WE_tccr0b & DI[4]) | (~WE_tccr0b & tsm & psr);

  reg t0_prev;
  wire t0_nops = (cs0 == 1);
  wire t0_clk =  (cs0 == 2) ? r_prescaler[2] // 1/8
               : (cs0 == 3) ? r_prescaler[5] // 1/64
               : (cs0 == 4) ? r_prescaler[7] // 1/256
               : (cs0 == 5) ? r_prescaler[9] // 1/1024
               : 0;
  wire t0_top_ocr0a = ((wgm0 == 2) | (wgm0 == 5) | (wgm0 == 7));
`ifdef TIM0_FUNC_MINIMUM
`else
  wire ocr0_chg_top = ((wgm0 == 1) | (wgm0 == 5));
  wire ocr0_chg_btm = ((wgm0 == 3) | (wgm0 == 7));
  wire ocr0_chg_imm = (wgm0[0] == 0);
`endif
`ifdef OUTPUT_MUX_TRI
  wire [7:0] i_DO = OE_ocr0b ? r_ocr0b : OE_ocr0a ? r_ocr0a : OE_tcnt0 ? r_tcnt0
              : OE_tccr0b ? i_tccr0b : OE_tccr0a ? i_tccr0a : 0 ;
  bufif1(DO,i_DO, (OE_ocr0b | OE_ocr0a | OE_tcnt0 | OE_tccr0b | OE_tccr0a) );
`else
  wire [7:0] i_DO = OE_ocr0b ? r_ocr0b : OE_ocr0a ? r_ocr0a : OE_tcnt0 ? r_tcnt0
              : OE_tccr0b ? i_tccr0b : OE_tccr0a ? i_tccr0a : 0 ;
  assign DO = i_DO;
`endif
  always @(negedge CLK)
  begin
    if (~RESET | v_psr) r_prescaler <= 1;
    else              r_prescaler <= r_prescaler + 10'h001;
  end

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_ocr0b <= 0; 	 
	r_ocr0a <= 0;
	cs0 <= 0;
	wgm0[2] <= 0;
	com0b <= 0;
	com0a <= 0;
	wgm0[1:0] <= 0;
	tsm <= 0;
	psr <= 0;
`ifdef TIM0_FUNC_MINIMUM
`else
	foc0b <= 0;
	foc0a <= 0;
`endif
	r_tcnt0 <= 0;
	r_tim0_ovf   <= 0;
	r_tim0_compa <= 0;
	r_tim0_compb <= 0;
	t0_prev <= 0;
      end
    else
      begin
	if(WE_ocr0b) r_ocr0b <= DI; 	 
	if(WE_ocr0a) r_ocr0a <= DI;
	if(WE_tccr0a) 
	begin
	  com0a <= DI[7:6];
	  com0b <= DI[5:4];
	  wgm0[1:0] <= DI[1:0];
	end
`ifdef ZZZ
	if(WE_tccr0b) psr <= DI[4];
	else if (~tsm) psr <= 0;
`else
	psr <= v_psr;
`endif

	if(WE_tccr0b) 
	begin
	  tsm <= DI[5];
	  wgm0[2] <= DI[3];
	  cs0 <= DI[2:0];
`ifdef TIM0_FUNC_MINIMUM
`else
	  foc0a <= DI[7];
	  foc0b <= DI[6];
`endif
	end
`ifdef TIM0_FUNC_MINIMUM
	if(WE_tcnt0) r_tcnt0 <= DI;
	else
	begin
	  if (t0_nops | (t0_clk ^ t0_prev) & t0_prev)
            begin
		if (t0_top_ocr0a & (r_tcnt0 == r_ocr0a)) r_tcnt0 <= 8'h00; 
		else 
		   begin
			r_tcnt0 <= r_tcnt0 + 8'h01;
 
			r_tim0_ovf   <= (r_tcnt0 + 8'h01 == 255);
			r_tim0_compa <= (r_tcnt0 + 8'h01 == r_ocr0a);
			r_tim0_compb <= (r_tcnt0 + 8'h01 == r_ocr0b);
		   end
	    end
	end
`else
NOT IMPLEMENTED YET
`endif

	t0_prev <= t0_clk;
      end
  end
endmodule

