`timescale 10ps / 10ps
`define GTK // use VCD

`include "rtavr_defs.v"

`define MODULE	tb_usart

module `MODULE;

   reg CLK;  // input
   reg RESET;  // input
   reg INT_ACK;  // input
`ifdef IOR_HAVE_TIM0
   wire IOR_INT0;  // input
`else
   reg IOR_INT0;  // input
`endif
`ifdef IOR_HAVE_TIM0
   wire IOR_TIM0_OVF;  // input
   wire IOR_TIM0_COMPA;  // input
   wire IOR_TIM0_COMPB;  // input
`else
   reg IOR_TIM0_OVF;  // input
   reg IOR_TIM0_COMPA;  // input
   reg IOR_TIM0_COMPB;  // input
`endif
`ifdef IOR_HAVE_SPI
   wire IOR_SPI_INT;  // input
`else
   reg IOR_SPI_INT;  // input
`endif
`ifdef IOR_HAVE_USART
   wire IOR_USART_TX;  // input
   wire IOR_USART_UDRE;  // input
   wire IOR_USART_RX;  // input
`else
   reg IOR_USART_TX;  // input
   reg IOR_USART_UDRE;  // input
   reg IOR_USART_RX;  // input
`endif

   wire [4:0] INT_VEC;

  reg [7:0] CLK_COUNT;  // input

  rtavr_intcont ICONT (.CLK(CLK), .RESET(RESET)

		, .INT_VEC(INT_VEC)
		, .INT_REQ(INT_REQ)
		, .INT_ACK(INT_ACK)

		, .IOR_INT0(IOR_INT0)
		, .IOR_RST_INT0(IOR_RST_INT0)

`ifdef IOR_HAVE_TIM0
		, .IOR_TIM0_OVF(IOR_TIM0_OVF)
		, .IOR_RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .IOR_TIM0_COMPA(IOR_TIM0_COMPA)
		, .IOR_RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .IOR_TIM0_COMPB(IOR_TIM0_COMPB)
		, .IOR_RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
`endif
`ifdef IOR_HAVE_SPI
		, .IOR_SPI_INT(IOR_SPI_INT)
		, .IOR_RST_SPI_INT(IOR_RST_SPI_INT)
`endif
		);
  reg IOR_RD;
  reg IOR_WR;
  reg [5:0] IOR_ADDR;
  reg [7:0] IOR_DI;
  wire [7:0] IOR_DO;
  inout [7:0] PORTC;
  reg miso;
//  assign PORTC[2]= PORTC[4];

  rtavr_ior IOR (.CLK(CLK), .RESET(RESET), .WR(IOR_WR) , .RD(IOR_RD)
		, .ADDR(IOR_ADDR[5:0]), .DI(IOR_DI[7:0]), .DO(IOR_DO[7:0])
  `ifdef IOR_HAVE_PORTA
		, .PORTA(PORTA[7:0])
  `endif
  `ifdef IOR_HAVE_PORTB
		, .PORTB(PORTB[7:0])
  `endif
  `ifdef IOR_HAVE_PORTC
		, .PORTC(PORTC[7:0])
  `endif
  `ifdef IOR_HAVE_SPI
		, .SPI_INT(IOR_SPI_INT)
		, .RST_SPI_INT(IOR_RST_SPI_INT)
  `endif
  `ifdef IOR_HAVE_USART
		, .USART_TX_INT(IOR_USART_TX)
		, .USART_UDRE_INT(IOR_USART_UDRE)
		, .USART_RX_INT(IOR_USART_RX)
		, .RST_USART_TX_INT(IOR_RST_USART_TX)
		, .RST_USART_UDRE_INT(IOR_RST_USART_UDRE)
		, .RST_USART_RX_INT(IOR_RST_USART_RX)
  `endif
  `ifdef IOR_HAVE_INT0
		, .INT0(IOR_INT0)
		, .RST_INT0(IOR_RST_INT0)
  `endif
  `ifdef IOR_HAVE_TIM0
		// INTERRUPTS 
		, .TIM0_OVF(IOR_TIM0_OVF)
		, .RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .TIM0_COMPA(IOR_TIM0_COMPA)
		, .RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .TIM0_COMPB(IOR_TIM0_COMPB)
		, .RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
  `endif
		);

   parameter CLKP = 5000;        //  50ns
   parameter CLKH = CLKP/2;      //
   parameter CLKL = CLKP-CLKH;   //
   
   always begin
      #CLKH CLK = 0;
      #CLKL CLK = 1;
   end

`ifdef GTK
   initial begin
      $dumpfile("tb.vcd");
      $dumpvars(0,`MODULE);
   end
`endif

   initial begin : test
      sys_reset();
      sys_wait(4);
//  parameter SS_BIT   = 16;      // PC0
//  parameter SCK_BIT  = 17;      // PC1
//  parameter MOSI_BIT = 20;      // PC4
//  parameter MISO_BIT = 18;      // PC2

      ior_out(6'h0e, 8'h02); // UBRRL 
      ior_out(6'h0f, 8'h00); // UBRRH 

      ior_out(6'h12, 8'h42); // UCSRA TXC=1 (CLR) , U2X = 1
      ior_out(6'h11, 8'h48); // UCSRB TXCIE = 1, TXEN = 1
      ior_out(6'h10, 8'h06); // UCSRC UPM=00 (PNONE) , USBS=0 (STOP 1)
			     // UCSZ=3 (8BIT) UCPOL=0
      sys_wait(4);
      ior_out(6'h13, 8'ha5); // UDR
      sys_wait_int(30);

      sys_wait(20);
      $display("*** Simulation Complete! ***");
      $finish;
   end

   task sys_reset;
      begin
	 INT_ACK = 0;
	 IOR_RD = 0;
	 IOR_WR = 0;
	 IOR_ADDR[5:0] = 0;
         IOR_DI[7:0] =0;
`ifdef IOR_HAVE_TIM0
`else
   	 IOR_INT0 = 0;
`endif
`ifdef IOR_HAVE_TIM0
`else
	 IOR_TIM0_OVF = 0;
	 IOR_TIM0_COMPA = 0;
	 IOR_TIM0_COMPB = 0;
`endif
         #1 RESET = 0;
         CLK = 0;
	 @(negedge CLK);
         #1 RESET = 1;
         @(posedge CLK);
      end
   endtask // sys_reset
   
   task ior_out;
      input [5:0] addr;
      input [7:0] data;
      begin
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = addr[5:0];
         IOR_DI[7:0] = data[7:0];
         IOR_WR = 1;
         @(posedge CLK);
	 @(negedge CLK);
	 #1 IOR_ADDR[5:0] = 0;
         IOR_DI[7:0] =0;
         IOR_WR = 0;
         @(posedge CLK);
      end
   endtask // ior_out

   task ior_in;
      input [5:0] addr;
      begin
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = addr[5:0];
         IOR_RD = 1;
         @(posedge CLK);
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = 0;
         IOR_RD = 0;
         @(posedge CLK);
      end
   endtask // ior_out

   task sys_wait;
      input [7:0] count;
      begin
        CLK_COUNT = 0;
        while ( CLK_COUNT < count )
        begin
           @(posedge CLK);
           CLK_COUNT = CLK_COUNT + 1;
        end
      end
   endtask // sys_wait

   task sys_wait_int;
      input [7:0] count;
      begin
        CLK_COUNT = 0;
        while ( ( CLK_COUNT < count ) & ~INT_REQ )
        begin
           @(posedge CLK);
           CLK_COUNT = CLK_COUNT + 1;
        end
        @(negedge CLK);
	INT_ACK = 1;
        @(negedge CLK);
        @(negedge CLK);
	INT_ACK = 0;
        @(posedge CLK);
      end
   endtask // sys_wait_int

   task sys_wait_rdy;
      input [7:0] count;
      begin
        CLK_COUNT = 0;
        while ( ( CLK_COUNT < count ) & ~INT_REQ )
        begin
           @(posedge CLK);
           CLK_COUNT = CLK_COUNT + 1;
        end
      end
   endtask // sys_wait_rdy

endmodule // test
